@@ -1,6 +1,6 @@
 ################################################################################
 #
-# Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
+# Copyright (C) 2019-2024 Advanced Micro Devices, Inc. All rights reserved.
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -29,7 +29,6 @@ from .SolutionStructs import Solution as OriginalSolution
 from .Utils import state, state_key_ordering
 
 from . import Common
-from . Common import globalParameters
 
 @state_key_ordering
 class FreeIndex:
@@ -318,7 +317,13 @@ class ProblemPredicate(Properties.Predicate):
         if key == "AssertSizeMultiple":
             return extractDimPredicate(cls, key, value, "SizeMultiple")
 
-        #Alpha and beta value assertions
+        # Arithmetic intensity assertions
+        if key == "AssertAIGreaterThanEqual":
+            return cls("AIGreaterThanEqual", value=value) if value > 0 else None
+        if key == "AssertAILessThanEqual":
+            return cls("AILessThanEqual", value=value) if value > 0 else None
+
+        # Alpha and beta value assertions
         if key == "AssertAlphaValue":
             return cls("AlphaValue", value=str(value)) if value != False else None
         if key == "AssertBetaValue":
@@ -349,9 +354,6 @@ class ProblemPredicate(Properties.Predicate):
 
             return cls(tag, index=index, value=value)
 
-        if key == "_WorkspaceSizePerElemC" and value > 0:
-            return cls("WorkspaceCheck", index=0, value=value)
-
         if key.startswith('Assert'):
             raise RuntimeError("Unknown assertion key: {}".format(key))
 
@@ -369,8 +371,8 @@ class ProblemPredicate(Properties.Predicate):
             # calculate the minimum supported free dimension size
             TLUA = state['ProblemType']['TLUA']
             TLUB = state['ProblemType']['TLUB']
-            minFree0 = state['GlobalLoadVectorWidthA'] if TLUA else 1
-            minFree1 = state['GlobalLoadVectorWidthB'] if TLUB else 1
+            minFree0 = max(state['GlobalLoadVectorWidthA'], 1) if TLUA else 1
+            minFree1 = max(state['GlobalLoadVectorWidthB'], 1) if TLUB else 1
             rv += [cls('LeadingFree0SizesGreaterOrEqual', value=minFree0)]
             rv += [cls('LeadingFree1SizesGreaterOrEqual', value=minFree1)]
 
@@ -384,6 +386,10 @@ class ProblemPredicate(Properties.Predicate):
             if ('_GlobalAccumulation' not in state) or (state['_GlobalAccumulation'] != 'MultipleBuffer'):
                 rv += [cls("DeterministicMode", value = False)]
 
+        if ('StreamK' in state) and (state['StreamK'] == 1):
+            # StreamK = 1 uses atomic for partial tiles
+            rv += [cls("DeterministicMode", value = False)]
+
         # debugging: mark this set to allow the problem always runnable with PK
         if 'PersistentKernel' in state and state['PersistentKernel']:
             rv += [cls("PersistentKernelCheck")]
@@ -422,8 +428,8 @@ class ProblemPredicate(Properties.Predicate):
                 val = min(val, state["AssertSizeLessThan"][1] - 1)
             rv += [cls('BufferStoreOffsetLimitCheck', value=val)]
 
-        if '_GlobalAccumulation' in state and state['_GlobalAccumulation'] != None:
-            value = globalParameters['MinKForGSU'] * state['GlobalSplitU']
+        if '_GlobalAccumulation' in state and state['_GlobalAccumulation'] != None and not state["StreamK"]:
+            value = state['MinKForGSU'] * state['GlobalSplitU']
             rv += [cls('GlobalSplitUCheckMinK', value=value)]
 
         return rv
@@ -437,6 +443,19 @@ class ProblemPredicate(Properties.Predicate):
         predicates = [p for p in map(cls.FromOriginalKeyPair, d.items()) if p is not None] + extraPreds
         return cls.And(predicates)
 
+class TaskPredicate(Properties.Predicate):
+    @classmethod
+    def FromOriginalKeyPair(cls, pair):
+        (key, value) = pair
+        if key == "_WorkspaceSizePerElemC" and value > 0:
+            return cls("WorkspaceCheck")
+        return None
+
+    @classmethod
+    def FromOriginalState(cls, d, problemType, morePreds=[]):
+        predicates = [p for p in map(cls.FromOriginalKeyPair, d.items()) if p is not None]
+        return cls.And(predicates)
+
 class SizeMapping:
     StateKeys = ['workGroup',
                  'macroTile',
@@ -449,11 +468,13 @@ class SizeMapping:
                  'packSummationDims',
                  'packBatchDims',
                  'magicDivAlg',
+                 'streamK',
                  'persistentKernel',
                  'persistentKernelAlongBatch',
                  'sourceKernel',
                  'globalAccumulation',
                  'workspaceSizePerElemC',
+                 'preloadKernargs'
                  ]
 
     @classmethod
@@ -463,6 +484,10 @@ class SizeMapping:
             globalAccum = 1
         if d['_GlobalAccumulation'] == 'MultipleBuffer':
             globalAccum = 2
+        if d['_GlobalAccumulation'] == 'PartialsBuffer':
+            globalAccum = 3
+
+        assert d["PreloadKernelArguments"] in (0, 1), d["PreloadKernelArguments"]
         return cls(workGroup             = d['WorkGroup'],
                    macroTile             = cls.ReadOriginalMacroTile(d),
                    threadTile            = d['ThreadTile'],
@@ -473,12 +498,14 @@ class SizeMapping:
                    staggerStrideShift    = d['_staggerStrideShift'] if '_staggerStrideShift' in d else 0,
                    packSummationDims     = d['PackSummationDims'] if 'PackSummationDims' in d else 0,
                    packBatchDims         = d['PackBatchDims'] if 'PackBatchDims' in d else 0,
+                   streamK               = d['StreamK'] if 'StreamK' in d else 0,
                    persistentKernel      = d['PersistentKernel'] if 'PersistentKernel' in d else 0,
                    persistentKernelAlongBatch   = d['PersistentKernelAlongBatch'] if 'PersistentKernelAlongBatch' in d else False,
                    magicDivAlg           = d.get('MagicDivAlg', 1),
                    sourceKernel          = d['KernelLanguage'] == 'Source',
                    globalAccumulation    = globalAccum,
                    workspaceSizePerElemC = d['_WorkspaceSizePerElemC'],
+                   preloadKernargs       = d["PreloadKernelArguments"]
                    )
 
     @classmethod
@@ -497,6 +524,7 @@ class Solution:
                 'problemType',
                 'hardwarePredicate',
                 'problemPredicate',
+                'taskPredicate',
                 'sizeMapping',
                 'debugKernel',
                 'libraryLogicIndex',
@@ -520,6 +548,7 @@ class Solution:
         rv.problemType = ProblemType.FromOriginalState(d['ProblemType'])
 
         rv.problemPredicate = ProblemPredicate.FromOriginalState(d, rv.problemType)
+        rv.taskPredicate = TaskPredicate.FromOriginalState(d, rv.problemType)
 
         if 'DebugKernel' in d:
             rv.debugKernel = d['DebugKernel']
@@ -562,6 +591,7 @@ class Solution:
         self.problemType = None
         self.hardwarePredicate = Hardware.HardwarePredicate('TruePred')
         self.problemPredicate = ProblemPredicate('TruePred')
+        self.taskPredicate = TaskPredicate('TruePred')
         self.sizeMapping = None
         self.debugKernel = False
         self.libraryLogicIndex = {}
