@@ -1,6 +1,6 @@
 /*******************************************************************************
  *
- * Copyright (C) 2016-2022 Advanced Micro Devices, Inc. All rights reserved.
+ * Copyright (C) 2016-2023 Advanced Micro Devices, Inc. All rights reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -174,6 +174,13 @@ constexpr __host__ __device__ bool operator==(tensile_complex<T> const& a,
     return (a.x == b.x) && (a.y == b.y);
 }
 
+template <typename T>
+constexpr __host__ __device__ bool operator!=(tensile_complex<T> const& a,
+                                              tensile_complex<T> const& b)
+{
+    return (a.x != b.x) || (a.y != b.y);
+}
+
 using tensile_float_complex  = tensile_complex<float>;
 using tensile_double_complex = tensile_complex<double>;
 
@@ -553,9 +560,7 @@ private:
 struct ProblemProperties
 {
 
-    // Constructor used in solution tables-
-    // See writeSolutionAndExactTable in TensileCreateLibrary - this constructor
-    // must be in-sync with the table written there.
+    // Constructor used in solution tables
     ProblemProperties(unsigned summationElementMultiple,
                       unsigned free0ElementMultiple,
                       unsigned free1ElementMultiple,
