@@ -2,7 +2,7 @@
  *
  * MIT License
  *
- * Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
+ * Copyright (C) 2019-2024 Advanced Micro Devices, Inc. All rights reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -35,6 +35,9 @@
 #include <cstdlib>
 #include <random>
 
+#include <iomanip>
+#include <regex>
+
 namespace Tensile
 {
     PerfModel perf;
@@ -255,7 +258,7 @@ namespace Tensile
                                                 TypedInputs const&                  inputs,
                                                 Hardware const&                     hardware) const
     {
-        TENSILE_ASSERT_EXC(sizeMapping.workGroupMapping >= 0);
+        //TENSILE_ASSERT_EXC(sizeMapping.workGroupMapping >= 0);
 
         TensorDescriptor const& a = problem.a();
         TensorDescriptor const& b = problem.b();
@@ -311,12 +314,25 @@ namespace Tensile
 
         rv.numWorkGroups.y *= sizeMapping.globalSplitU;
 
-        if(sizeMapping.persistentKernel != 0)
+        size_t cuCount = 0;
+        size_t skGrid  = 0;
+        auto   tiles   = problem.getNumTiles(sizeMapping);
+        if(sizeMapping.streamK != 0 || sizeMapping.persistentKernel != 0)
         {
             AMDGPU const* pAMDGPU = dynamic_cast<AMDGPU const*>(&hardware);
             assert(pAMDGPU != nullptr && pAMDGPU->computeUnitCount != 0);
+            cuCount = pAMDGPU->computeUnitCount;
+            if(sizeMapping.streamK != 0)
+            {
+                skGrid             = getSKGrid(hardware, tiles);
+                rv.numWorkGroups.x = skGrid;
+                rv.numWorkGroups.y = 1;
+                rv.numWorkGroups.z = 1;
+            }
+        }
 
-            size_t cuCount       = pAMDGPU->computeUnitCount;
+        if(sizeMapping.persistentKernel != 0)
+        {
             size_t finalPKValue  = sizeMapping.persistentKernel;
             size_t problemGroups = rv.numWorkGroups.x * rv.numWorkGroups.y;
             if(sizeMapping.persistentKernelAlongBatch)
@@ -351,6 +367,12 @@ namespace Tensile
 
         rv.sharedMemBytes = 0;
 
+        size_t startStrideCD = problemType.useInitialStridesCD ? 0 : 1;
+        size_t startStrideAB = problemType.useInitialStridesAB ? 0 : 1;
+
+        auto problemSizes = problem.problemSizes();
+        TENSILE_ASSERT_EXC(problemSizes.size() > 0);
+
         if(!isSourceKernel())
         {
             uint64_t tensor2dSizeC = c.totalAllocatedElements();
@@ -361,20 +383,53 @@ namespace Tensile
                                          ? b.totalAllocatedElements()
                                          : problem.allocatedElementsNonBatchB();
 
-            rv.args.append<uint64_t>("tensor2dSizeC", tensor2dSizeC);
-            rv.args.append<uint64_t>("tensor2dSizeA", tensor2dSizeA);
-            rv.args.append<uint64_t>("tensor2dSizeB", tensor2dSizeB);
+            if(sizeMapping.preloadKernargs)
+            {
+                TENSILE_ASSERT_EXC(problemType.stridedBatched);
+                TENSILE_ASSERT_EXC(problemType.useBeta);
+                TENSILE_ASSERT_EXC(startStrideAB == 1);
+
+                rv.args.append<uint64_t>("tensor2dSizeA", tensor2dSizeA);
+                rv.args.append<uint64_t>("tensor2dSizeB", tensor2dSizeB);
+
+                auto aptr = inputs.a + a.offset();
+                auto bptr = inputs.b + b.offset();
+                rv.args.append<typename TypedInputs::AType const*>("a", aptr);
+                rv.args.append<typename TypedInputs::BType const*>("b", bptr);
+
+                rv.args.append<typename TypedInputs::AlphaType>("alpha", inputs.alpha);
+                if(std::is_same<typename TypedInputs::AlphaType, Half>::value && !isSourceKernel())
+                    rv.args.append<typename TypedInputs::AlphaType>("alpha_2", inputs.alpha);
+
+                for(size_t i = startStrideAB; i < a.dimensions(); i++)
+                    rv.args.append<uint32_t>(concatenate_if<T_Debug>("strideA", i), a.strides()[i]);
+
+                for(size_t i = startStrideAB; i < b.dimensions(); i++)
+                    rv.args.append<uint32_t>(concatenate_if<T_Debug>("strideB", i), b.strides()[i]);
+
+                rv.args.append<uint32_t>("sizeSum", problemSizes.back());
+            }
+
+            // rv.args.append<uint64_t>("tensor2dSizeC", tensor2dSizeC);
+
+            if(!sizeMapping.preloadKernargs)
+            {
+                rv.args.append<uint64_t>("tensor2dSizeA", tensor2dSizeA);
+                rv.args.append<uint64_t>("tensor2dSizeB", tensor2dSizeB);
+            }
         }
 
-        if(sizeMapping.globalAccumulation)
+        if(sizeMapping.globalAccumulation && sizeMapping.streamK < 2)
         {
             rv.args.append<void const*>("ws_d", inputs.ws);
             rv.args.append<void const*>("ws_c", inputs.ws);
         }
         else if(problemType.stridedBatched)
         {
-            rv.args.append<typename TypedInputs::DType const*>("d", inputs.d);
-            rv.args.append<typename TypedInputs::CType const*>("c", inputs.c);
+            auto dptr = inputs.d + d.offset();
+            auto cptr = inputs.c + c.offset();
+            rv.args.append<typename TypedInputs::DType const*>("d", dptr);
+            rv.args.append<typename TypedInputs::CType const*>("c", cptr);
         }
         else
         {
@@ -384,8 +439,13 @@ namespace Tensile
 
         if(problemType.stridedBatched)
         {
-            rv.args.append<typename TypedInputs::AType const*>("a", inputs.a);
-            rv.args.append<typename TypedInputs::BType const*>("b", inputs.b);
+            if(!sizeMapping.preloadKernargs)
+            {
+                auto aptr = inputs.a + a.offset();
+                auto bptr = inputs.b + b.offset();
+                rv.args.append<typename TypedInputs::AType const*>("a", aptr);
+                rv.args.append<typename TypedInputs::BType const*>("b", bptr);
+            }
         }
         else
         {
@@ -393,14 +453,30 @@ namespace Tensile
             rv.args.append<typename TypedInputs::BType const* const*>("batchB", inputs.batchB);
         }
 
-        rv.args.append<uint64_t>("offsetD", d.offset());
-        rv.args.append<uint64_t>("offsetC", c.offset());
-        rv.args.append<uint64_t>("offsetA", a.offset());
-        rv.args.append<uint64_t>("offsetB", b.offset());
+        if(sizeMapping.streamK >= 2)
+        {
+            // StreamK workspace + flags
+            rv.args.append<void const*>("ws", inputs.ws);
+            void*  ws          = inputs.ws;
+            size_t flagsOffset = partialTileSize(skGrid);
+            void*  flags       = (void*)(static_cast<char*>(ws) + flagsOffset);
+            rv.args.append<void*>("Flags", flags);
+        }
 
-        rv.args.append<typename TypedInputs::AlphaType>("alpha", inputs.alpha);
-        if(std::is_same<typename TypedInputs::AlphaType, Half>::value && !isSourceKernel())
-            rv.args.append<typename TypedInputs::AlphaType>("alpha_2", inputs.alpha);
+        if(!problemType.stridedBatched)
+        {
+            rv.args.append<uint64_t>("offsetD", d.offset());
+            rv.args.append<uint64_t>("offsetC", c.offset());
+            rv.args.append<uint64_t>("offsetA", a.offset());
+            rv.args.append<uint64_t>("offsetB", b.offset());
+        }
+
+        if(!sizeMapping.preloadKernargs)
+        {
+            rv.args.append<typename TypedInputs::AlphaType>("alpha", inputs.alpha);
+            if(std::is_same<typename TypedInputs::AlphaType, Half>::value && !isSourceKernel())
+                rv.args.append<typename TypedInputs::AlphaType>("alpha_2", inputs.alpha);
+        }
 
         if(problemType.useBeta)
         {
@@ -409,10 +485,7 @@ namespace Tensile
                 rv.args.append<typename TypedInputs::BetaType>("beta_2", inputs.beta);
         }
 
-        size_t startStrideCD = problemType.useInitialStridesCD ? 0 : 1;
-        size_t startStrideAB = problemType.useInitialStridesAB ? 0 : 1;
-
-        if(sizeMapping.globalAccumulation)
+        if(sizeMapping.globalAccumulation && sizeMapping.streamK < 2)
         {
             size_t wsStride = startStrideCD ? d.sizes()[0] : 1;
             for(size_t i = startStrideCD; i < d.dimensions(); i++)
@@ -437,18 +510,22 @@ namespace Tensile
                 rv.args.append<uint32_t>(concatenate_if<T_Debug>("strideC", i), c.strides()[i]);
         }
 
-        for(size_t i = startStrideAB; i < a.dimensions(); i++)
-            rv.args.append<uint32_t>(concatenate_if<T_Debug>("strideA", i), a.strides()[i]);
+        if(!sizeMapping.preloadKernargs)
+        {
+            for(size_t i = startStrideAB; i < a.dimensions(); i++)
+                rv.args.append<uint32_t>(concatenate_if<T_Debug>("strideA", i), a.strides()[i]);
 
-        for(size_t i = startStrideAB; i < b.dimensions(); i++)
-            rv.args.append<uint32_t>(concatenate_if<T_Debug>("strideB", i), b.strides()[i]);
+            for(size_t i = startStrideAB; i < b.dimensions(); i++)
+                rv.args.append<uint32_t>(concatenate_if<T_Debug>("strideB", i), b.strides()[i]);
+        }
 
         {
-            int idx = 0;
-            for(auto size : problem.problemSizes())
+            int maxIdx = problemSizes.size();
+            if(sizeMapping.preloadKernargs)
+                maxIdx--;
+            for(int idx = 0; idx < maxIdx; idx++)
             {
-                rv.args.append<uint32_t>(concatenate_if<T_Debug>("size_", idx), size);
-                idx++;
+                rv.args.append<uint32_t>(concatenate_if<T_Debug>("size_", idx), problemSizes[idx]);
             }
         }
 
@@ -539,7 +616,8 @@ namespace Tensile
             }
         }
 
-        rv.args.append<int32_t>("staggerUIter", staggerUIter(problem, inputs, hardware));
+        if(sizeMapping.staggerU)
+            rv.args.append<int32_t>("staggerUIter", staggerUIter(problem, inputs, hardware));
 
         rv.args.append<uint32_t>("problemNumGroupTiles0", problemNumGroupTiles0);
         rv.args.append<uint32_t>("problemNumGroupTiles1", problemNumGroupTiles1);
@@ -551,15 +629,20 @@ namespace Tensile
             uint32_t magicNumberWgmRemainder1 = 0;
 
             // conditional args, aligned with KernelWriterAssembly.py
-            if(sizeMapping.persistentKernel != 0)
+            if(sizeMapping.persistentKernel != 0 || sizeMapping.streamK != 0)
             {
                 uint32_t magicShift;
                 rv.args.append<uint32_t>("magicNumberProblemNumGroupTiles0",
                                          magicNumber(2, problemNumGroupTiles0, &magicShift));
                 rv.args.append<uint32_t>("magicShiftProblemNumGroupTiles0", magicShift);
+            }
+
+            if(sizeMapping.persistentKernel != 0)
+            {
                 rv.args.append<uint32_t>("gridNumWorkGroups0", rv.numWorkGroups.x);
             }
 
+            // if(sizeMapping.persistentKernelAlongBatch || sizeMapping.streamK != 0)
             if(sizeMapping.persistentKernelAlongBatch)
             {
                 uint32_t numGroupTiles0x1 = problemNumGroupTiles0 * problemNumGroupTiles1;
@@ -571,18 +654,83 @@ namespace Tensile
                 rv.args.append<uint32_t>("magicShiftProblemNumGroupTiles0By1", magicShift);
             }
 
+            if(sizeMapping.streamK != 0)
+            {
+                auto     itersPerTile = problem.getItersPerTile(sizeMapping);
+                auto     totalIters   = tiles * itersPerTile;
+                uint32_t magicNumberItersPerTile;
+                uint32_t magicShiftItersPerTile;
+                magicNumberItersPerTile = magicNumber(2, itersPerTile, &magicShiftItersPerTile);
+
+                rv.args.append<uint32_t>("itersPerTile", itersPerTile);
+                rv.args.append<uint32_t>("magicNumberItersPerTile", magicNumberItersPerTile);
+                rv.args.append<uint32_t>("magicShiftItersPerTile", magicShiftItersPerTile);
+
+                uint32_t numGroupTiles0x1 = problemNumGroupTiles0 * problemNumGroupTiles1;
+                uint32_t magicNumProblemNumGroupTiles0By1;
+                uint32_t magicShiftProblemNumGroupTiles0By1;
+                magicNumProblemNumGroupTiles0By1
+                    = magicNumber(2, numGroupTiles0x1, &magicShiftProblemNumGroupTiles0By1);
+                rv.args.append<uint32_t>("magicNumProblemNumGroupTiles0By1",
+                                         magicNumProblemNumGroupTiles0By1);
+                rv.args.append<uint32_t>("magicShiftProblemNumGroupTiles0By1",
+                                         magicShiftProblemNumGroupTiles0By1);
+
+                rv.args.append<uint32_t>("totalIters", totalIters);
+                if(sizeMapping.streamK < 3) // Basic SK
+                {
+                    uint32_t itersPerWave = CeilDivide(totalIters, rv.numWorkGroups.x);
+                    rv.args.append<uint32_t>("SKItersPerWG", itersPerWave);
+                }
+                else if(sizeMapping.streamK == 3) // Two-tile SK
+                {
+                    bool bigEnough = tiles > skGrid;
+                    // skTiles is number of Stream-K tiles to complete
+                    // Two-tile algorithm causes each WG to run an even number of Stream-K iterations,
+                    // followed by an even number of data-parllel tiles.
+                    // If total tiles is evenly divisble by grid size,
+                    // then no Stream-K tiles are needed, all data-parallel
+                    uint32_t skTiles = skGrid;
+                    if(tiles % skGrid != 0)
+                    {
+                        // Number of data-parallel tiles on each workgroup would be:
+                        // dpTilesPerWG = bigEnough ? (tiles - skTiles) / skGrid : 0;
+                        skTiles = bigEnough ? skGrid + tiles % skGrid : tiles;
+                    }
+
+                    uint32_t skItersPerWG = skTiles * itersPerTile / skGrid;
+                    uint32_t skExtraIters = skTiles * itersPerTile % (skGrid);
+
+                    rv.args.append<uint32_t>("SKItersPerWG", skItersPerWG);
+                    rv.args.append<uint32_t>("skGrid", skGrid);
+                    rv.args.append<uint32_t>("skTiles", skTiles);
+                    rv.args.append<uint32_t>("skExtraIters", skExtraIters);
+                }
+            }
+
             if(sizeMapping.workGroupMapping != 0)
             {
-                numFullBlocks = problemNumGroupTiles1 / sizeMapping.workGroupMapping;
-                wgmRemainder1 = problemNumGroupTiles1 % sizeMapping.workGroupMapping;
+                if(sizeMapping.workGroupMapping >= 0)
+                {
+                    numFullBlocks = problemNumGroupTiles1 / sizeMapping.workGroupMapping;
+                    wgmRemainder1 = problemNumGroupTiles1 % sizeMapping.workGroupMapping;
+                }
+                else
+                {
+                    numFullBlocks = problemNumGroupTiles0 / (-sizeMapping.workGroupMapping);
+                    wgmRemainder1 = problemNumGroupTiles0 % (-sizeMapping.workGroupMapping);
+                }
                 if(wgmRemainder1 == 0)
                     wgmRemainder1 = sizeMapping.workGroupMapping;
                 magicNumberWgmRemainder1 = smallMagicNumber(wgmRemainder1);
             }
 
-            rv.args.append<uint32_t>("numFullBlocks", numFullBlocks);
-            rv.args.append<uint32_t>("wgmRemainder1", wgmRemainder1);
-            rv.args.append<uint32_t>("magicNumberWgmRemainder1", magicNumberWgmRemainder1);
+            if(std::abs(sizeMapping.workGroupMapping) > 1)
+            {
+                rv.args.append<uint32_t>("numFullBlocks", numFullBlocks);
+                rv.args.append<uint32_t>("wgmRemainder1", wgmRemainder1);
+                rv.args.append<uint32_t>("magicNumberWgmRemainder1", magicNumberWgmRemainder1);
+            }
         }
 
         if(problemType.stochasticRounding)
@@ -611,6 +759,64 @@ namespace Tensile
         return sizeMapping.sourceKernel;
     }
 
+    template <typename TypedInputs, bool T_Debug>
+    KernelInvocation ContractionSolution::generateStreamKInitCall(Problem const&     problem,
+                                                                  TypedInputs const& inputs,
+                                                                  Hardware const&    hardware) const
+    {
+        TensorDescriptor const& c = problem.c();
+        TensorDescriptor const& d = problem.d();
+
+        KernelInvocation rv;
+
+        rv.args = KernelArguments(T_Debug);
+
+        rv.args.reserve(512, 64);
+
+        rv.kernelName = streamKInitKernelName(problem, inputs, hardware);
+
+        rv.workGroupSize.x = 256;
+        rv.workGroupSize.y = 1;
+        rv.workGroupSize.z = 1;
+
+        auto   tiles  = problem.getNumTiles(sizeMapping);
+        size_t skGrid = getSKGrid(hardware, tiles);
+        size_t wiZ    = 1;
+        for(size_t i = 0; i < problem.batchIndices().size(); i++)
+            wiZ *= problem.batchSize(i);
+        size_t flagCount = skGrid * wiZ;
+
+        rv.numWorkGroups.x = CeilDivide(flagCount, rv.workGroupSize.x);
+        rv.numWorkGroups.y = 1;
+        rv.numWorkGroups.z = 1;
+
+        rv.numWorkItems.x = rv.workGroupSize.x * rv.numWorkGroups.x;
+        rv.numWorkItems.y = rv.workGroupSize.y * rv.numWorkGroups.y;
+        rv.numWorkItems.z = rv.workGroupSize.z * rv.numWorkGroups.z;
+
+        void*  ws          = inputs.ws;
+        size_t flagsOffset = partialTileSize(skGrid);
+        void*  flags       = (void*)(static_cast<char*>(ws) + flagsOffset);
+        rv.args.append<void*>("Flags", flags);
+
+        rv.args.append<uint32_t>("flagCount", flagCount);
+
+        //Pass along code object dependency
+        // TODO check this
+        rv.codeObjectFile = codeObjectFilename.load();
+
+        return rv;
+    }
+
+    template <typename TypedInputs>
+    std::string ContractionSolution::streamKInitKernelName(Problem const&     problem,
+                                                           TypedInputs const& inputs,
+                                                           Hardware const&    hardware) const
+    {
+        std::string name = "WSFlags";
+        return name;
+    }
+
     template <typename TypedInputs, bool T_Debug>
     KernelInvocation ContractionSolution::generateBetaOnlyCall(Problem const&     problem,
                                                                TypedInputs const& inputs,
@@ -652,17 +858,28 @@ namespace Tensile
         if(sizeMapping.globalAccumulation)
             rv.args.append<void*>("WS", inputs.ws);
         else if(problemType.stridedBatched)
-            rv.args.append<typename TypedInputs::DType*>("D", inputs.d);
+        {
+            auto dptr = inputs.d;
+            dptr += d.offset();
+            rv.args.append<typename TypedInputs::DType*>("D", dptr);
+        }
         else
             rv.args.append<typename TypedInputs::DType const* const*>("batchD", inputs.batchD);
 
         if(problemType.stridedBatched)
-            rv.args.append<typename TypedInputs::CType const*>("C", inputs.c);
+        {
+            auto cptr = inputs.c;
+            cptr += c.offset();
+            rv.args.append<typename TypedInputs::CType const*>("C", cptr);
+        }
         else
             rv.args.append<typename TypedInputs::CType const* const*>("batchC", inputs.batchC);
 
-        rv.args.append<uint64_t>("offsetD", d.offset());
-        rv.args.append<uint64_t>("offsetC", c.offset());
+        if(!problemType.stridedBatched)
+        {
+            rv.args.append<uint64_t>("offsetD", d.offset());
+            rv.args.append<uint64_t>("offsetC", c.offset());
+        }
 
         if(sizeMapping.globalAccumulation)
         {
@@ -725,6 +942,10 @@ namespace Tensile
     KernelInvocation ContractionSolution::generateOutputConversionCall(
         Problem const& problem, TypedInputs const& inputs, Hardware const& hardware) const
     {
+        AMDGPU const* pAMDGPU = dynamic_cast<AMDGPU const*>(&hardware);
+        assert(pAMDGPU != nullptr && pAMDGPU->computeUnitCount != 0);
+        size_t cuCount = pAMDGPU->computeUnitCount;
+
         TensorDescriptor const& c = problem.c();
         TensorDescriptor const& d = problem.d();
 
@@ -734,8 +955,6 @@ namespace Tensile
 
         rv.args.reserve(512, 64);
 
-        rv.kernelName = outputConversionKernelName(problem, inputs, hardware);
-
         rv.workGroupSize.x = 256;
         rv.workGroupSize.y = 1;
         rv.workGroupSize.z = 1;
@@ -750,7 +969,64 @@ namespace Tensile
         for(size_t i = 0; i < problem.batchIndices().size(); i++)
             wiZ *= problem.batchSize(i);
 
-        rv.numWorkGroups.x = CeilDivide(wiX * wiY * wiZ, rv.workGroupSize.x);
+        const unsigned int numThreadsPerCU = 256;
+        unsigned int       gsu             = static_cast<int>(sizeMapping.globalSplitU);
+        if(sizeMapping.globalAccumulation == 1)
+            // globalAccumulation = 1 case, ignore globalSplitU and use 1
+            gsu = 1;
+        // wider global load for postGSU
+        // only for compute type = Float
+        bool supportedTypeForVWopt
+            = problem.alphaType() == DataType::Float || problem.alphaType() == DataType::Double;
+        size_t total = wiX * wiY * wiZ;
+        int    vw    = 1;
+        size_t threshVW2
+            = cuCount * numThreadsPerCU * 2; // should be more than number of physical threads * vw
+        if(supportedTypeForVWopt && total > threshVW2 && problem.freeSizeA(0) % 2 == 0)
+            vw = 2;
+
+        // parallel reduction width calculation
+        // only for compute type = Float
+        bool supportedTypeForReductionOpt
+            = problem.alphaType() == DataType::Float || problem.alphaType() == DataType::Double;
+        size_t threshReduction
+            = cuCount
+              * numThreadsPerCU; // should be less than number of physical threads / reduction
+        const unsigned int maxReductionConst = 4;
+        const unsigned int minGSUperReduction
+            = 32; // Minimum GSU=128 for Reduction=4, GSU=64 for Reduction2
+        unsigned int maxReduction = std::min(maxReductionConst, gsu / minGSUperReduction);
+        unsigned int reduction    = 1;
+        if(supportedTypeForReductionOpt && (gsu & (gsu - 1)) == 0 && maxReduction > 1)
+        {
+            // apply reduction only if float compute type and gsu is power of 2 (for small array only)
+            reduction = maxReduction;
+            while(reduction > 1)
+            {
+                size_t totalThread = total * reduction / vw;
+                if(gsu % reduction == 0 && totalThread <= threshReduction)
+                    // found an applicable reduction
+                    break;
+                // for next loop
+                reduction /= 2;
+            }
+            if(reduction == 2 && problem.alphaType() == DataType::Float)
+                // so far, reduction=2 does not perform well
+                reduction = 1;
+        }
+
+        // GSU loop unroll opt
+        // only for compute type = Float, Double
+        bool supportedTypeForUnrollOpt
+            = problem.alphaType() == DataType::Float || problem.alphaType() == DataType::Double;
+        int gsuUnrollUnit = 16 * reduction; // must match Tensile generator code
+        if(supportedTypeForUnrollOpt == false || sizeMapping.globalAccumulation == 1)
+            gsuUnrollUnit = 1;
+
+        rv.kernelName = outputConversionKernelName(
+            problem, inputs, gsu, vw, reduction, gsuUnrollUnit, hardware);
+
+        rv.numWorkGroups.x = CeilDivide(wiX * wiY * wiZ * reduction, rv.workGroupSize.x * vw);
         rv.numWorkGroups.y = 1;
         rv.numWorkGroups.z = 1;
 
@@ -759,19 +1035,30 @@ namespace Tensile
         rv.numWorkItems.z = rv.workGroupSize.z * rv.numWorkGroups.z;
 
         if(problemType.stridedBatched)
-            rv.args.append<typename TypedInputs::DType*>("D", inputs.d);
+        {
+            auto dptr = inputs.d;
+            dptr += d.offset();
+            rv.args.append<typename TypedInputs::DType*>("D", dptr);
+        }
         else
             rv.args.append<typename TypedInputs::DType const* const*>("batchD", inputs.batchD);
 
         rv.args.append<void*>("WS", inputs.ws);
 
         if(problemType.stridedBatched)
-            rv.args.append<typename TypedInputs::CType const*>("C", inputs.c);
+        {
+            auto cptr = inputs.c;
+            cptr += c.offset();
+            rv.args.append<typename TypedInputs::CType const*>("C", cptr);
+        }
         else
             rv.args.append<typename TypedInputs::CType const* const*>("batchC", inputs.batchC);
 
-        rv.args.append<uint64_t>("offsetD", d.offset());
-        rv.args.append<uint64_t>("offsetC", c.offset());
+        if(!problemType.stridedBatched)
+        {
+            rv.args.append<uint64_t>("offsetD", d.offset());
+            rv.args.append<uint64_t>("offsetC", c.offset());
+        }
 
         if(sizeMapping.globalAccumulation == 2)
             rv.args.append<typename TypedInputs::AlphaType>("alpha", inputs.alpha);
@@ -803,10 +1090,7 @@ namespace Tensile
             idx++;
         }
 
-        if(sizeMapping.globalAccumulation == 1)
-            rv.args.append<uint32_t>("gsu", 1);
-        else
-            rv.args.append<uint32_t>("gsu", sizeMapping.globalSplitU);
+        rv.args.append<uint32_t>("gsu", gsu);
 
         if(problemType.stochasticRounding)
         {
@@ -827,7 +1111,17 @@ namespace Tensile
 
     bool ContractionSolution::canSolve(Problem const& problem, Hardware const& hardware) const
     {
-        return (*problemPredicate)(problem) && (*hardwarePredicate)(hardware);
+        static const bool debug = Debug::Instance().printPredicateEvaluation();
+
+        Task task(hardware, problem, *this);
+        if(debug)
+        {
+            hardwarePredicate->debugEval(hardware, std::cout);
+            problemPredicate->debugEval(problem, std::cout);
+            taskPredicate->debugEval(task, std::cout);
+        }
+        return (*taskPredicate)(task) && (*problemPredicate)(problem)
+               && (*hardwarePredicate)(hardware);
     }
 
     bool ContractionSolution::matchesProblemType(Problem const&  problem,
@@ -849,6 +1143,10 @@ namespace Tensile
     template <typename TypedInputs>
     std::string ContractionSolution::outputConversionKernelName(Problem const&     problem,
                                                                 TypedInputs const& inputs,
+                                                                int                gsu,
+                                                                int                vw,
+                                                                int                reduction,
+                                                                int                gsuUnrollUnit,
                                                                 Hardware const&    hardware) const
     {
         std::string name = concatenate(
@@ -861,6 +1159,27 @@ namespace Tensile
 
         name += "_PostGSU";
 
+        // add extra string for gsu
+        // This part must match tensile code generation (in KernelWriterConversion.py)
+        // add mod related str (only for gsuUnrollUnit > 1)
+        if(gsuUnrollUnit > 1)
+        {
+            size_t gsuMod = gsu % gsuUnrollUnit;
+            if(gsuMod == 0)
+                gsuMod = gsuUnrollUnit;
+            std::string modStr = "";
+            if(gsu > gsuUnrollUnit)
+            {
+                modStr = "_mod" + std::to_string(gsuUnrollUnit);
+            }
+            name += std::to_string(gsuMod) + modStr;
+        }
+        if(vw >= 2)
+            name += "_VW" + std::to_string(vw);
+
+        if(reduction >= 2)
+            name += "_R" + std::to_string(reduction);
+
         return name;
     }
 
@@ -917,7 +1236,17 @@ namespace Tensile
 
         std::vector<KernelInvocation> rv;
 
-        if(sizeMapping.globalSplitU > 1 && sizeMapping.globalAccumulation != 2)
+        if(sizeMapping.streamK >= 2)
+        {
+            if(debug)
+                rv.push_back(generateStreamKInitCall<TypedInputs, true>(problem, inputs, hardware));
+            else
+                rv.push_back(
+                    generateStreamKInitCall<TypedInputs, false>(problem, inputs, hardware));
+        }
+
+        if(sizeMapping.streamK == 1
+           || (sizeMapping.globalSplitU > 1 && sizeMapping.globalAccumulation != 2))
         {
             if(debug)
                 rv.push_back(generateBetaOnlyCall<TypedInputs, true>(problem, inputs, hardware));
@@ -930,8 +1259,9 @@ namespace Tensile
         else
             rv.push_back(generateSingleCall<TypedInputs, false>(problem, inputs, hardware));
 
-        if(sizeMapping.globalAccumulation)
+        if(sizeMapping.globalAccumulation && !sizeMapping.streamK)
         {
+            // TODO Streamk May need conversion call for HPA??
             if(debug)
                 rv.push_back(
                     generateOutputConversionCall<TypedInputs, true>(problem, inputs, hardware));
@@ -943,6 +1273,55 @@ namespace Tensile
         return rv;
     }
 
+    bool ContractionSolution::getMatrixInstructionFromKernelName(
+        vector4<std::uint32_t>& matrixInst) const
+    {
+
+        std::string regexp_string("_MI(\\d+)x(\\d+)x(\\d+)x(\\d+)_");
+        std::regex  miRegex(regexp_string);
+        std::smatch matches;
+
+        std::string kName = this->KernelName();
+
+        bool matched
+            = std::regex_search(kName, matches, miRegex)
+              && (matches.size()
+                  == 5 /* 1 (index 0) for entire match and 4 for the sub-experssion matches */);
+
+        if(matched)
+        {
+            // the first sub_match element (index 0) corresponds to the entire match
+            matrixInst.x = atoi(matches[1].str().c_str());
+            matrixInst.y = atoi(matches[2].str().c_str());
+            matrixInst.z = atoi(matches[3].str().c_str());
+            matrixInst.w = atoi(matches[4].str().c_str());
+        }
+
+        return matched;
+    }
+
+    bool ContractionSolution::getGSUAlgorithmFromKernelName(std::string& gsuAlg) const
+    {
+        std::string regexp_string("_GSUA([S|M])B_");
+        std::regex  miRegex(regexp_string);
+        std::smatch matches;
+
+        std::string kName = this->KernelName();
+
+        bool matched
+            = std::regex_search(kName, matches, miRegex)
+              && (matches.size()
+                  == 2 /* 1 (index 0) for entire match and 1 for the sub-experssion matches */);
+
+        if(matched)
+        {
+            // the first sub_match element (index 0) corresponds to the entire match
+            gsuAlg = (matches[1].compare("S") == 0) ? "SingleBuffer" : "MultipleBuffer";
+        }
+
+        return matched;
+    }
+
     std::vector<KernelInvocation>
         ContractionSolution::solve(ContractionSolution::Problem const& problem,
                                    ContractionSolution::Inputs const&  inputs,
@@ -951,6 +1330,22 @@ namespace Tensile
         if(Debug::Instance().printWinningKernelName())
             std::cout << "Running kernel: " << this->KernelName() << std::endl;
 
+        if(Debug::Instance().printKernelCommonParams())
+        {
+            std::cout << "Kernel name: " << this->KernelName() << std::endl;
+            std::cout << "Kernel parameters: " << std::endl;
+
+            vector4<std::uint32_t> matrixInst;
+            if(this->getMatrixInstructionFromKernelName(matrixInst))
+                std::cout << std::right << std::setw(30) << "MatrixInstruction: " << matrixInst
+                          << std::endl;
+
+            std::string GSUAlg;
+            if(this->getGSUAlgorithmFromKernelName(GSUAlg))
+                std::cout << std::right << std::setw(30) << "GSUAlgorithm: " << GSUAlg << std::endl;
+
+            std::cout << this->sizeMapping << std::endl;
+        }
         // retreive alpha/beta type set via setAlpha/BetaType()
         auto alphaType = problem.alphaType();
         auto betaType  = problem.betaType();
@@ -1174,11 +1569,52 @@ namespace Tensile
         return spm;
     }
 
-    size_t ContractionSolution::requiredWorkspaceSize(Problem const& problem) const
+    size_t ContractionSolution::requiredWorkspaceSize(Problem const&  problem,
+                                                      Hardware const& hardware) const
     {
         size_t size = 0;
 
-        size += problem.d().totalLogicalElements() * sizeMapping.workspaceSizePerElemC;
+        if(sizeMapping.streamK >= 2)
+        {
+            auto   tiles  = problem.getNumTiles(sizeMapping);
+            size_t skGrid = getSKGrid(hardware, tiles);
+            // Get space required for partial tiles
+            size += partialTileSize(skGrid);
+            // Add space for flags
+            // Flags for partial tiles - dword per flag for fast addressing and comparisons
+            size += skGrid * 4;
+            // size *= batches; // TODO need tile and flag per batch
+        }
+        else
+            size += problem.d().totalLogicalElements() * sizeMapping.workspaceSizePerElemC;
+
+        return size;
+    }
+
+    size_t ContractionSolution::getSKGrid(Hardware const& hardware, size_t tiles) const
+    {
+        AMDGPU const* pAMDGPU = dynamic_cast<AMDGPU const*>(&hardware);
+        assert(pAMDGPU != nullptr && pAMDGPU->computeUnitCount != 0);
+        size_t cuCount = pAMDGPU->computeUnitCount;
+        size_t skGrid  = cuCount;
+        if(pAMDGPU->skMaxCUs > 0)
+            skGrid = min(skGrid, pAMDGPU->skMaxCUs);
+        if(pAMDGPU->skDynamicGrid)
+            skGrid = min(skGrid, tiles);
+        if(pAMDGPU->skGridMultiplier > 1)
+            skGrid = skGrid * pAMDGPU->skGridMultiplier;
+        return skGrid;
+    }
+
+    size_t ContractionSolution::partialTileSize(size_t skGrid) const
+    {
+        size_t size = 0;
+
+        size_t tileSize
+            = sizeMapping.macroTile.x * sizeMapping.macroTile.y * sizeMapping.workspaceSizePerElemC;
+        size += tileSize * skGrid; // Partials tile per WG
+        // TODO batches
+        // TODO round up for alignment?
 
         return size;
     }
@@ -1468,4 +1904,41 @@ namespace Tensile
                       << " shiftPtrElemB=" << st.shiftPtrElemB << " depthUorMT0=" << st.depthUorMT0
                       << " depthUorMT1=" << st.depthUorMT1;
     }
+
+    std::ostream& operator<<(std::ostream& stream, const SizeMapping& sizeMapping)
+    {
+        std::ios_base::fmtflags flags(stream.flags());
+
+        stream << std::right << std::setw(30) << "workGroupSize: " << sizeMapping.workGroupSize
+               << std::endl
+               << std::setw(30) << "threadTile: " << sizeMapping.threadTile << std::endl
+               << std::setw(30) << "macroTile: " << sizeMapping.macroTile << std::endl
+
+               << std::setw(30) << "staggerU: " << sizeMapping.staggerU << std::endl
+               << std::setw(30) << "depthU: " << sizeMapping.depthU << std::endl
+               << std::setw(30) << "globalSplitU: " << sizeMapping.globalSplitU << std::endl
+               << std::setw(30) << "staggerStrideShift: " << sizeMapping.staggerStrideShift
+               << std::endl
+               << std::setw(30) << "workGroupMapping: " << sizeMapping.workGroupMapping << std::endl
+
+               << std::setw(30) << "packBatchDims: " << sizeMapping.packBatchDims << std::endl
+               << std::setw(30) << "packSummationDims: " << sizeMapping.packSummationDims
+               << std::endl
+               << std::setw(30) << "magicDivAlg: " << sizeMapping.magicDivAlg << std::endl
+               << std::setw(30) << "streamK: " << sizeMapping.streamK << std::endl
+               << std::setw(30) << "persistentKernel: " << sizeMapping.persistentKernel << std::endl
+               << std::setw(30)
+               << "persistentKernelAlongBatch: " << sizeMapping.persistentKernelAlongBatch
+               << std::endl
+
+               << std::setw(30) << "sourceKernel: " << sizeMapping.sourceKernel << std::endl
+               << std::setw(30) << "globalAccumulation: " << sizeMapping.globalAccumulation
+               << std::endl
+               << std::setw(30) << "workspaceSizePerElemC: " << sizeMapping.workspaceSizePerElemC
+               << std::endl;
+
+        stream.flags(flags);
+
+        return stream;
+    }
 } // namespace Tensile
