@@ -1,6 +1,6 @@
 ################################################################################
 #
-# Copyright (C) 2016-2022 Advanced Micro Devices, Inc. All rights reserved.
+# Copyright (C) 2016-2023 Advanced Micro Devices, Inc. All rights reserved.
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -39,7 +39,6 @@ from .KernelWriterAssembly import KernelWriterAssembly
 from .KernelWriterSource import KernelWriterSource
 from .SolutionLibrary import MasterSolutionLibrary
 from .SolutionStructs import Solution
-from .SolutionWriter import SolutionWriter
 
 import argparse
 import collections
@@ -149,7 +148,10 @@ def getAssemblyCodeObjectFiles(kernels, kernelWriterAssembly, outputPath):
     return coFiles
 
 def which(p):
-    exes = [p+x for x in ['.bat', '', '.exe']]  # bat may be front end for file with no extension
+    if os.name == "nt":
+        exes = [p+x for x in ['.bat', '', '.exe']]  # bat may be front end for file with no extension
+    else:
+        exes = [p+x for x in ['', '.exe', '.bat']]
     system_path = os.environ['PATH'].split(os.pathsep)
     if p == 'hipcc' and 'CMAKE_CXX_COMPILER' in os.environ and os.path.isfile(os.environ['CMAKE_CXX_COMPILER']):
         return os.environ['CMAKE_CXX_COMPILER']
@@ -249,8 +251,11 @@ def buildSourceCodeObjectFile(CxxCompiler, outputPath, kernelFile):
         outflag = "-output"
 
       infile = os.path.join(buildPath, objectFilename)
+      bundler = globalParameters["ClangOffloadBundlerPath"]
+      if bundler is None:
+        raise ValueError('No bundler available; set TENSILE_ROCM_OFFLOAD_BUNDLER_PATH to point to clang-offload-bundler.')
       try:
-        bundlerArgs = [globalParameters["ClangOffloadBundlerPath"], "-type=o", "%s=%s" % (inflag, infile), "-list"]
+        bundlerArgs = [bundler, "-type=o", "%s=%s" % (inflag, infile), "-list"]
         listing = subprocess.check_output(bundlerArgs, stderr=subprocess.STDOUT).decode().split("\n")
         for target in listing:
           matched = re.search("gfx.*$", target)
@@ -271,8 +276,8 @@ def buildSourceCodeObjectFile(CxxCompiler, outputPath, kernelFile):
             #Compilation
             if outfile:
               coFilenames.append(os.path.split(outfile)[1])
-              #bundlerArgs = [globalParameters["ClangOffloadBundlerPath"], "-type=o", "-targets=%s" % target, "-inputs=%s" % infile, "-outputs=%s" % outfile, "-unbundle"]
-              bundlerArgs = [globalParameters["ClangOffloadBundlerPath"], "-type=o", "-targets=%s" % target,
+              #bundlerArgs = [bundler, "-type=o", "-targets=%s" % target, "-inputs=%s" % infile, "-outputs=%s" % outfile, "-unbundle"]
+              bundlerArgs = [bundler, "-type=o", "-targets=%s" % target,
                            "%s=%s" % (inflag, infile), "%s=%s" % (outflag, outfile), "-unbundle"]
               if globalParameters["PrintCodeCommands"]:
                 print(' '.join(bundlerArgs))
@@ -285,8 +290,8 @@ def buildSourceCodeObjectFile(CxxCompiler, outputPath, kernelFile):
         for i in range(len(archs)):
           outfile = os.path.join(buildPath, "{0}-000-{1}.hsaco".format(soFilename, archs[i]))
           coFilenames.append(os.path.split(outfile)[1])
-          #bundlerArgs = [globalParameters["ClangOffloadBundlerPath"], "-type=o", "-targets=hip-amdgcn-amd-amdhsa--%s" % cmdlineArchs[i], "-inputs=%s" % infile, "-outputs=%s" % outfile, "-unbundle"]
-          bundlerArgs = [globalParameters["ClangOffloadBundlerPath"], "-type=o", "-targets=hip-amdgcn-amd-amdhsa--%s" % cmdlineArchs[i],
+          #bundlerArgs = [bundler, "-type=o", "-targets=hip-amdgcn-amd-amdhsa--%s" % cmdlineArchs[i], "-inputs=%s" % infile, "-outputs=%s" % outfile, "-unbundle"]
+          bundlerArgs = [bundler, "-type=o", "-targets=hip-amdgcn-amd-amdhsa--%s" % cmdlineArchs[i],
                          "%s=%s" % (inflag, infile), "%s=%s" % (outflag, outfile), "-unbundle"]
           if globalParameters["PrintCodeCommands"]:
             print(' '.join(bundlerArgs))
@@ -474,8 +479,8 @@ def buildKernelSourceAndHeaderFiles(results, outputPath, kernelsWithBuildErrs):
 ################################################################################
 # Write Solutions and Kernels for BenchmarkClient or LibraryClient
 ################################################################################
-def writeSolutionsAndKernels(outputPath, CxxCompiler, problemTypes, solutions, kernels, kernelHelperObjs, \
-    solutionWriter, kernelWriterSource, kernelWriterAssembly, errorTolerant=False):
+def writeKernels(outputPath, CxxCompiler, problemTypes, solutions, kernels, kernelHelperObjs, \
+    kernelWriterSource, kernelWriterAssembly, errorTolerant=False):
   start = time.time()
 
   codeObjectFiles = []
@@ -616,160 +621,20 @@ def writeSolutionsAndKernels(outputPath, CxxCompiler, problemTypes, solutions, k
 
   return codeObjectFiles
 
-def writeSolutionAndExactTable(scheduleName, deviceNames, schedProbName, problemType, \
-                               solutionsForSchedule, solutionNames, exactLogic):
-  s = ""
-  s += "namespace { // Start schedule '%s'\n" % scheduleName
-
-  s += "// solution table - function, name, assertion requirements\n"
-  s += "static const SolutionInfo solutionTable_%s[] = {\n" % (schedProbName)
-  for i in range(0, len(solutionsForSchedule)):
-    solution = solutionsForSchedule[i]
-    solutionName = solutionNames[i]
-    s += "  {(void*)%s, \"%s\", {%d, %d, %d, %d, %d, %d, %d} }%s // %d" % \
-      (solutionName, solutionName, \
-        solution["AssertSummationElementMultiple"], \
-        solution["AssertFree0ElementMultiple"], \
-        solution["AssertFree1ElementMultiple"], \
-        solution["AssertMinApproxSize"], \
-        False, \
-        solution["PackBatchDims"]==2, \
-        solution["PackBatchDims"]==1, \
-        "," if i < len(solutionsForSchedule)-1 else "", \
-        i)
-    s += "\n"
-
-  s += "};\n\n"
-
-  # Write the exact problems here
-  s += "// table of exact problem dims and selected solutionIdx\n"
-  s += "static const std::pair<const ProblemKey_%s, int> embeddedExactTable_%s[] = {\n" % (problemType,schedProbName)
-  numSizes = problemType["TotalIndices"]
-  for ruleIdx in range(0, len(exactLogic)):
-    rule = exactLogic[ruleIdx]
-    problemSize = rule[0][:numSizes]
-    solutionIdx = rule[1][0]
-    solutionGFlops = rule[1][1]
-    s += " { {"
-    for i in range(0, len(problemSize)):
-      if i == 0:
-        s += "%u" % problemSize[i]
-      else:
-        s += ", %u" % problemSize[i]
-    s += "}, %u}" % (solutionIdx)
-    s += "," if ruleIdx != len(exactLogic)-1 else " "
-    s += " // %.0f GFlop/s" % (solutionGFlops)
-    s += "\n"
-  s += "};\n\n"
-
-  # Create a solution mapper and init with the table above:
-  s += "// The solution master constructor here adds device to the master solution mapper\n"
-  s += "// The entrypoint to find a solution for this problem is through the master solution master\n"
-  s += "static SolutionMapper_%s solutionMapper_%s(\n" % (problemType, schedProbName)
-  s += "  \"%s\", // schedule+problem name\n" % (schedProbName)
-  s += "  solutionTable_%s, %u,\n" % (schedProbName, len(solutionsForSchedule))
-  s += "  embeddedExactTable_%s, %u,\n" % (schedProbName, len(exactLogic))
-  s += "  &problemType_%s);\n" % (problemType)
-
-  s += "} // end anonymous namespace\n"
-  return s
-
-
-################################################################################
-# Write Range Logic Recursive
-# ptr :
-#   True : write logic to return the function pointer
-#   False : write logic to return the function name
-################################################################################
-def writeExactLogic(problemType, indexOrder,
-                    solutionsForSchedule, exactLogic, \
-                    solutionNames, ptr):
-  s = ""
-  s += "  ProblemDims_%s pdims(" % problemType
-  indexChars = globalParameters["IndexChars"]
-  firstStrideAB = 0 if problemType["UseInitialStridesAB"] else 1
-  firstStrideCD = 0 if problemType["UseInitialStridesCD"] else 1
-  lastStrideD = problemType["NumIndicesC"]
-  lastStrideC = problemType["NumIndicesC"]
-  lastStrideA = len(problemType["IndexAssignmentsA"])
-  lastStrideB = len(problemType["IndexAssignmentsB"])
-  for i in range(firstStrideCD,lastStrideD):
-    if i != firstStrideCD: s += ", "
-    s += "strideD%u%s" % (i, indexChars[i])
-  for i in range(firstStrideCD,lastStrideC):
-    s += ", strideC%u%s" % (i, indexChars[i])
-  for i in range(firstStrideAB,lastStrideA):
-    s += ", strideA%u%s" % (i, \
-        indexChars[problemType["IndexAssignmentsA"][i]])
-  for i in range(firstStrideAB,lastStrideB):
-    s += ", strideB%u%s" % (i, \
-        indexChars[problemType["IndexAssignmentsB"][i]])
-  for i in range(0,len(indexOrder)):
-    s += ", size%s" % indexChars[i]
-  s += ");\n"
-
-  s += "  auto solutionMapper = reinterpret_cast<SolutionMapper_%s *> (masterSolutionMapper_%s.mapper());\n"  \
-      % (problemType, problemType)
-  if ptr:
-    s += "  return solutionMapper->getSolutionWithFallback(pdims,&masterSolutionMapper_%s);\n" % problemType
-  else:
-    s += "  return solutionMapper->getSolutionWithFallback(pdims,&masterSolutionMapper_%s)->_info->_name;\n" % problemType
-
-  return s
-
-
-################################################################################
-# Write Solution Call
-################################################################################
-def writeSolutionCall(solutionName, problemType):
-  indexChars = globalParameters["IndexChars"]
-  s = ""
-  s += "%s(" % solutionName
-  # solution parameters
-  s += " dataD, dataC, dataA, dataB, alpha"
-  if problemType["UseBeta"]:
-    s += ", beta"
-  s += ", offsetC, offsetA, offsetB"
-  firstStrideAB = firstStrideCD = 1
-  if problemType["UseInitialStridesAB"]:
-    firstStrideAB = 0
-  if problemType["UseInitialStridesCD"]:
-    firstStrideCD = 0
-  lastStrideD = problemType["NumIndicesC"]
-  lastStrideC = problemType["NumIndicesC"]
-  lastStrideA = len(problemType["IndexAssignmentsA"])
-  lastStrideB = len(problemType["IndexAssignmentsB"])
-  for i in range(firstStrideCD,lastStrideD):
-    s += ", strideD%u%s" % (i, indexChars[i])
-  for i in range(firstStrideCD,lastStrideC):
-    s += ", strideC%u%s" % (i, indexChars[i])
-  for i in range(firstStrideAB,lastStrideA):
-    s += ", strideA%u%s" % (i, \
-        indexChars[problemType["IndexAssignmentsA"][i]])
-  for i in range(firstStrideAB,lastStrideB):
-    s += ", strideB%u%s" % (i, \
-        indexChars[problemType["IndexAssignmentsB"][i]])
-  for i in range(0, problemType["TotalIndices"]):
-    s += ", size%s" % indexChars[i]
-  s += ", stream, numInputEvents, inputEvents, outputEvent )"
-  return s
-
 ##############################################################################
 # Min Naming / Solution and Kernel Writers
 ##############################################################################
-def getSolutionAndKernelWriters(solutions, kernels):
+def getKernelWriters(solutions, kernels):
 
   # if any kernels are assembly, append every ISA supported
-  solutionSerialNaming = Solution.getSerialNaming(solutions)
   kernelSerialNaming   = Solution.getSerialNaming(kernels)
 
   solutionMinNaming    = Solution.getMinNaming(solutions)
   kernelMinNaming      = Solution.getMinNaming(kernels)
-  solutionWriter       = SolutionWriter(solutionMinNaming, solutionSerialNaming, kernelMinNaming, kernelSerialNaming)
   kernelWriterSource   = KernelWriterSource(kernelMinNaming, kernelSerialNaming)
   kernelWriterAssembly = KernelWriterAssembly(kernelMinNaming, kernelSerialNaming)
 
-  return (solutionWriter, kernelWriterSource, kernelWriterAssembly, kernelMinNaming, solutionMinNaming)
+  return (kernelWriterSource, kernelWriterAssembly, kernelMinNaming, solutionMinNaming)
 
 ################################################################################
 # copy static cpp files and headers
@@ -791,7 +656,7 @@ def copyStaticFiles(outputPath=None):
 
   return libraryStaticFiles
 
-def buildObjectFileNames(solutionWriter, kernelWriterSource, kernelWriterAssembly, solutions, kernels, kernelHelperObjs):
+def buildObjectFileNames(kernelWriterSource, kernelWriterAssembly, solutions, kernels, kernelHelperObjs):
 
   # Build lists of output object names
   sourceKernelNames = []
@@ -1046,8 +911,7 @@ def generateLogicDataAndSolutions(logicFiles, args):
   masterLibraries = {}
   fullMasterLibrary = None
 
-  nextSolIndex = 0
-
+  nextSolIndex = {}
   for logic in Utils.tqdm(libraries, "Processing logic data"):
     (_, architectureName, _, solutionsForSchedule, _, newLibrary) = logic
 
@@ -1058,10 +922,14 @@ def generateLogicDataAndSolutions(logicFiles, args):
         masterLibraries[architectureName] = deepcopy(newLibrary)
         masterLibraries[architectureName].version = args.version
     elif globalParameters["SeparateArchitectures"] or globalParameters["LazyLibraryLoading"]:
+
       if architectureName in masterLibraries:
-        nextSolIndex = masterLibraries[architectureName].merge(deepcopy(newLibrary), nextSolIndex)
+        nextSolIndex[architectureName] = masterLibraries[architectureName].merge(deepcopy(newLibrary), nextSolIndex[architectureName])
       else:
         masterLibraries[architectureName] = deepcopy(newLibrary)
+        archIndexMap = MasterSolutionLibrary.ArchitectureIndexMap(architectureName)
+        masterLibraries[architectureName].remapSolutionIndicesStartingFrom(archIndexMap)
+        nextSolIndex[architectureName] = archIndexMap
         masterLibraries[architectureName].version = args.version
     else:
       if fullMasterLibrary is None:
@@ -1079,8 +947,7 @@ def generateLogicDataAndSolutions(logicFiles, args):
     if "fallback" in masterLibraries.keys():
       for key, value in masterLibraries.items():
         if key != "fallback":
-          value.merge(deepcopy(masterLibraries["fallback"]))
-
+          value.insert(deepcopy(masterLibraries["fallback"]))
       masterLibraries.pop("fallback")
 
     for _, masterLibrary in masterLibraries.items():
@@ -1107,14 +974,14 @@ def writeBenchmarkClientFiles(libraryWorkingPath, tensileSourcePath, solutions,
       copyStaticFiles(libraryWorkingPath)
 
   kernels, kernelsBetaOnly, _ = generateKernelObjectsFromSolutions(solutions)
-  solutionWriter, kernelWriterSource, kernelWriterAssembly, \
-    kernelMinNaming, _ = getSolutionAndKernelWriters(solutions, kernels)
+  kernelWriterSource, kernelWriterAssembly, \
+    kernelMinNaming, _ = getKernelWriters(solutions, kernels)
 
   # write solution, kernels and CMake
   problemType = solutions[0]["ProblemType"]
-  codeObjectFiles = writeSolutionsAndKernels( \
+  codeObjectFiles = writeKernels( \
     libraryWorkingPath, cxxCompiler, [problemType], solutions, kernels, kernelsBetaOnly, \
-    solutionWriter, kernelWriterSource, kernelWriterAssembly, errorTolerant=True )
+    kernelWriterSource, kernelWriterAssembly, errorTolerant=True )
 
   newLibraryDir = ensurePath(os.path.join(libraryWorkingPath, 'library'))
   newLibraryFile = os.path.join(newLibraryDir, "TensileLibrary.yaml")
@@ -1152,6 +1019,23 @@ def WriteClientLibraryFromSolutions(solutionList, libraryWorkingPath, tensileSou
 
   return (codeObjectFiles, newLibrary)
 
+################################################################################
+# Write Master Solution Index CSV
+################################################################################
+def writeMasterSolutionIndexCSV(outputPath, masterLibraries):
+  libraryPath = os.path.join(outputPath, "library")
+  ensurePath(libraryPath)
+  try:
+    with open(os.path.join(libraryPath, "TensileMasterSolutionIndex.csv"), "w") as indexFile:
+      indexFile.write("architectureName,libraryName,libraryIndex,solutionIndex,solutionName\n")
+      for arch,lib in masterLibraries.items():
+        for lazylibname,lazylibvals in lib.lazyLibraries.items():
+          for solidx,solution in lazylibvals.solutions.items():
+            line = ",".join(str(x) for x in [arch, lazylibname, solidx, solution.index, solution.name])
+            indexFile.write("%s\n" %(line))
+  except IOError as err:
+    print1("Error writing MasterSolutionIndex %s" % err)
+
 ################################################################################
 # Tensile Create Library
 ################################################################################
@@ -1219,6 +1103,8 @@ def TensileCreateLibrary():
   argParser.add_argument("--global-parameters", nargs="+", type=splitExtraParameters, default=[])
   argParser.add_argument("--ignore-asm-cap-cache", dest="IgnoreAsmCapCache", action="store_true", default=False,
                          help="Ignore asm cap cache and derive the asm caps at runtime")
+  argParser.add_argument("--write-master-solution-index",   dest="WriteMasterSolutionIndex", action="store_true",
+                          default=False, help="Output master solution index in csv format.")
   args = argParser.parse_args()
 
   logicPath = args.LogicPath
@@ -1258,7 +1144,8 @@ def TensileCreateLibrary():
   arguments["CpuThreads"] = args.CpuThreads
   arguments["PrintLevel"] = args.PrintLevel
   arguments["IgnoreAsmCapCache"] = args.IgnoreAsmCapCache
-  
+  arguments["WriteMasterSolutionIndex"] = args.WriteMasterSolutionIndex
+
   for key, value in args.global_parameters:
     arguments[key] = value
 
@@ -1294,6 +1181,10 @@ def TensileCreateLibrary():
                        and (any(logicArch in os.path.splitext(f)[0] for logicArch in logicArchs) \
                        or "hip" in os.path.splitext(f)[0]) ]
 
+  # Skip experimental libraries (if exists) when building without lazy loading
+  if not globalParameters["LazyLibraryLoading"]:
+    logicFiles = [f for f in logicFiles if not globalParameters["ExperimentalLogicDir"] in f]
+  
   print1("# LibraryLogicFiles:" % logicFiles)
   for logicFile in logicFiles:
     print1("#   %s" % logicFile)
@@ -1305,11 +1196,14 @@ def TensileCreateLibrary():
   # Parse logicData, solutions, and masterLibraries from logic files
   solutions, masterLibraries, fullMasterLibrary = generateLogicDataAndSolutions(logicFiles, args)
 
+  if globalParameters["LazyLibraryLoading"] and arguments["WriteMasterSolutionIndex"]:
+    writeMasterSolutionIndexCSV(outputPath, masterLibraries)
+
   kernels, kernelHelperObjs, _ = generateKernelObjectsFromSolutions(solutions)
 
   # if any kernels are assembly, append every ISA supported
-  solutionWriter, kernelWriterSource, kernelWriterAssembly, \
-    kernelMinNaming, _ = getSolutionAndKernelWriters(solutions, kernels)
+  kernelWriterSource, kernelWriterAssembly, \
+    kernelMinNaming, _ = getKernelWriters(solutions, kernels)
 
   staticFiles = copyStaticFiles(outputPath)
 
@@ -1318,7 +1212,7 @@ def TensileCreateLibrary():
    sourceKernelFiles,
    asmKernelFiles,
    sourceLibFiles,
-   asmLibFiles) = buildObjectFileNames(solutionWriter, kernelWriterSource, \
+   asmLibFiles) = buildObjectFileNames(kernelWriterSource, \
     kernelWriterAssembly, solutions, kernels, kernelHelperObjs)
 
   (_,
@@ -1352,8 +1246,8 @@ def TensileCreateLibrary():
       outputPath )
 
   # write solutions and kernels
-  codeObjectFiles = writeSolutionsAndKernels(outputPath, CxxCompiler, None, solutions,
-                                             kernels, kernelHelperObjs, solutionWriter, kernelWriterSource, kernelWriterAssembly)
+  codeObjectFiles = writeKernels(outputPath, CxxCompiler, None, solutions,
+                                             kernels, kernelHelperObjs, kernelWriterSource, kernelWriterAssembly)
 
   bothLibSet = set(sourceLibPaths + asmLibPaths)
   setA = set( map( os.path.normcase, set(codeObjectFiles) ) )
@@ -1407,8 +1301,8 @@ def TensileCreateLibrary():
     theMasterLibrary = list(masterLibraries.values())[0]
 
   if args.EmbedLibrary is not None:
-      embedFileName = os.path.join(outputPath, "library/{}.cpp".format(args.EmbedLibrary))
-      with EmbeddedData.EmbeddedDataFile(embedFileName) as embedFile:
+      embedFileNameTemp = os.path.join(outputPath, "library/{}.temp".format(args.EmbedLibrary))
+      with EmbeddedData.EmbeddedDataFile(embedFileNameTemp) as embedFile:
 
           ext = ".yaml" if globalParameters["LibraryFormat"] == "yaml" else ".dat"
           embedFile.embed_file(theMasterLibrary.cpp_base_class, masterFile + ext, nullTerminated=True,
@@ -1417,6 +1311,8 @@ def TensileCreateLibrary():
           for co in Utils.tqdm(codeObjectFiles):
               embedFile.embed_file("SolutionAdapter", co, nullTerminated=False,
                                    key=args.EmbedLibraryKey)
+      embedFileNameCpp = os.path.join(outputPath, "library/{}.cpp".format(args.EmbedLibrary))
+      os.rename(embedFileNameTemp, embedFileNameCpp)
 
   if args.BuildClient:
     print1("# Building Tensile Client")
